#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul  7 19:22:12 2020

@author: carlosesteveyague
"""

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import time

#define the shape of the environment (i.e., the states)
environment_rows = 11
environment_columns = 11

#Create a 5D numpy array to hold the current Q-values for each state and action pair: Q(s, a) 
#The array contains 11 rows and 11 columns (to match the shape of the environment), 11 rows and 11 columns for the state of the opponent
#and 5 layers for each of the possible actions.
#The "action" dimension consists of 5 layers that will allow us to keep track of the Q-values for each possible action in
#each state. 
#The value of each (state, action) pair is initialized to 0.
q_Hvalues = np.zeros((environment_rows, environment_columns, environment_rows, environment_columns, 5))

#The value function para the evader. Same as fo the Hunter but with only 4 possible actions.
q_Gvalues = np.zeros((environment_rows, environment_columns, environment_rows, environment_columns, 4))

#define actions
#numeric action codes: 0 = up, 1 = right, 2 = down, 3 = left, 4 = stay
Hactions = ['up', 'right', 'down', 'left','stay']
Gactions = ['up', 'right', 'down', 'left']

##Uncomment the following line to freeze the Hunter
#Hactions = ['stay', 'stay', 'stay', 'stay','stay']

#Create a 2D numpy array to hold the rewards for each state. 
#The array contains 11 rows and 11 columns (to match the shape of the environment), and each value is initialized to -100.
rewards = np.full((environment_rows, environment_columns), -100.)
#set the reward for the target area to 100



#define green area for rows 1 through 9
aisles = {} #store locations in a dictionary

for i in range(1,environment_rows//3):
    aisles[i] = [i for i in range(1, environment_columns-1)]
    
for i in range(environment_rows//3, 2*environment_rows//3):
    aisles[i] = [i for i in range(1, environment_columns//3)]
    aisles[i].append([i for i in range(2*environment_columns//3+1, environment_columns-1)])
   
for i in range(2*environment_rows//3,environment_rows-1):
    aisles[i] = [i for i in range(1, environment_columns-1)]


#set the rewards for all green squares
for row_index in range(1, environment_rows-1):
  for column_index in aisles[row_index]:
    rewards[row_index, column_index] = -1
    
#set the rewards for the case the hunter catches the evader
#and when the evader reaches the goal
def get_reward(H_row, H_column, G_row, G_column):
    if H_row ==G_row and H_column == G_column:
        return 100, -100
    elif G_row== environment_rows-1 and G_column == environment_columns//2:
        return -100, 100
    else:
        return rewards[H_row, H_column], rewards[G_row, G_column]
  
  
#define a function that determines if the specified location is a terminal state
def is_terminal_state(current_Hrow_index, current_Hcolumn_index, current_Grow_index, current_Gcolumn_index):
  #if the reward for this location is -1, then it is not a terminal state (i.e., it is a 'green square')
  if get_reward(current_Hrow_index, current_Hcolumn_index, current_Grow_index, current_Gcolumn_index)[0] ==-1 and get_reward(current_Hrow_index, current_Hcolumn_index, current_Grow_index, current_Gcolumn_index)[1] ==-1:
    return False
  else:
    return True

#define a function that will choose a random, non-terminal starting location
def get_starting_location():
  current_Hrow_index = np.random.randint(environment_rows)
  current_Grow_index = np.random.randint(environment_rows)
  current_Hcolumn_index = np.random.randint(environment_columns)
  current_Gcolumn_index = np.random.randint(1, environment_columns-1)
  #continue choosing random row and column indexes until a non-terminal state is identified
  while is_terminal_state(current_Hrow_index, current_Hcolumn_index, current_Grow_index, current_Gcolumn_index):
    current_Hrow_index = np.random.randint(environment_rows)
    current_Hcolumn_index = np.random.randint(environment_columns)
    current_Grow_index = np.random.randint(environment_rows)
    current_Gcolumn_index = np.random.randint(1, environment_columns-1)
  return current_Hrow_index, current_Hcolumn_index, current_Grow_index, current_Gcolumn_index

#define an epsilon greedy algorithm that will choose which action to take next (i.e., where to move next)
def get_next_Haction(current_Hrow_index, current_Hcolumn_index, current_Grow_index, current_Gcolumn_index, epsilonH):
  #if a randomly chosen value between 0 and 1 is less than epsilon, 
  #then choose the most promising value from the Q-table for this state.
  if np.random.random() < epsilonH:
      return np.argmax(q_Hvalues[current_Hrow_index, current_Hcolumn_index, current_Grow_index, current_Gcolumn_index])
  else: #choose a random action
      return np.random.randint(5)

def get_next_Gaction(current_Hrow_index, current_Hcolumn_index, current_Grow_index, current_Gcolumn_index, epsilonG):
  #if a randomly chosen value between 0 and 1 is less than epsilon, 
  #then choose the most promising value from the Q-table for this state.
  if np.random.random() < epsilonG:
      return np.argmax(q_Gvalues[current_Hrow_index, current_Hcolumn_index, current_Grow_index, current_Gcolumn_index])
  else: #choose a random action
      return np.random.randint(4)
  
    
#define a function that will get the next location based on the chosen action
def get_next_location(current_Hrow_index, current_Hcolumn_index, current_Grow_index, current_Gcolumn_index, Haction_index, Gaction_index):
  new_Hrow_index = current_Hrow_index
  new_Hcolumn_index = current_Hcolumn_index
  new_Grow_index = current_Grow_index
  new_Gcolumn_index = current_Gcolumn_index
  if Hactions[Haction_index] == 'up' and current_Hrow_index > 0:
    new_Hrow_index -= 1
  elif Hactions[Haction_index] == 'right' and current_Hcolumn_index < environment_columns - 1:
    new_Hcolumn_index += 1
  elif Hactions[Haction_index] == 'down' and current_Hrow_index < environment_rows - 1:
    new_Hrow_index += 1
  elif Hactions[Haction_index] == 'left' and current_Hcolumn_index > 0:
    new_Hcolumn_index -= 1
    
  if Gactions[Gaction_index] == 'left' and current_Gcolumn_index > 0:
    new_Gcolumn_index -= 1
  elif Gactions[Gaction_index] == 'right' and current_Gcolumn_index < environment_columns - 1:
    new_Gcolumn_index += 1
  elif Gactions[Gaction_index] == 'down' and current_Grow_index < environment_rows - 1:
    new_Grow_index += 1
  elif Gactions[Gaction_index] == 'up' and current_Grow_index > 0:
    new_Grow_index -= 1
  return new_Hrow_index, new_Hcolumn_index, new_Grow_index, new_Gcolumn_index


  
def start_hunting(start_Hrow_index, start_Hcolumn_index, start_Grow_index, start_Gcolumn_index,T, epsH, epsG):
  #return immediately if this is an invalid starting location
  if is_terminal_state(start_Hrow_index, start_Hcolumn_index, start_Grow_index, start_Gcolumn_index):
    return [], 0
  else: #if this is a 'legal' starting location
    current_Hrow_index, current_Hcolumn_index = start_Hrow_index, start_Hcolumn_index
    current_Grow_index, current_Gcolumn_index = start_Grow_index, start_Gcolumn_index
    shortest_path = []
    shortest_path.append([current_Hrow_index, current_Hcolumn_index, current_Grow_index, current_Gcolumn_index])
    t=0
    #continue moving along the path until we reach the goal (i.e., the item packaging location)
    while not is_terminal_state(current_Hrow_index, current_Hcolumn_index, current_Grow_index, current_Gcolumn_index) and t<T:
      #get the best action to take
      Haction_index = get_next_Haction(current_Hrow_index, current_Hcolumn_index, current_Grow_index, current_Gcolumn_index, epsH)
      Gaction_index = get_next_Gaction(current_Hrow_index, current_Hcolumn_index, current_Grow_index, current_Gcolumn_index, epsG)
      #move to the next location on the path, and add the new location to the list
      current_Hrow_index, current_Hcolumn_index, current_Grow_index, current_Gcolumn_index = get_next_location(current_Hrow_index, current_Hcolumn_index, current_Grow_index, current_Gcolumn_index, Haction_index, Gaction_index)
      shortest_path.append([current_Hrow_index, current_Hcolumn_index, current_Grow_index, current_Gcolumn_index])
      t+=1
    return shortest_path, t

#define training parameters
epsilonH = 0.9 #the percentage of time when we should take the best action (instead of a random action)
epsilonG = 0.9

discount_factor = 0.9 #discount factor for future rewards
learning_rate = 0.9 #the rate at which the AI agent should learn
episodes = 3000 #number of experiments during the training



def training(episodes):
    
    #run through 1000 training episodes
    for episode in range(episodes):
        #get the starting location for this episode
        row_Hindex, column_Hindex, row_Gindex, column_Gindex = get_starting_location()
        
        #continue taking actions (i.e., moving) until we reach a terminal state
        #(i.e., until we reach the item packaging area or crash into an item storage location)
        while not is_terminal_state(row_Hindex, column_Hindex, row_Gindex, column_Gindex):
            #choose which action to take (i.e., where to move next)
            Haction_index = get_next_Haction(row_Hindex, column_Hindex, row_Gindex, column_Gindex, epsilonH)
            Gaction_index = get_next_Gaction(row_Hindex, column_Hindex, row_Gindex, column_Gindex, epsilonG)
            
            #perform the chosen action, and transition to the next state (i.e., move to the next location)
            old_Hrow_index, old_Hcolumn_index = row_Hindex, column_Hindex #store the old row and column indexes
            old_Grow_index, old_Gcolumn_index = row_Gindex, column_Gindex
            row_Hindex, column_Hindex, row_Gindex, column_Gindex = get_next_location(row_Hindex, column_Hindex, row_Gindex, column_Gindex, Haction_index, Gaction_index)
            
            #receive the reward for moving to the new state, and calculate the temporal difference
            Hreward, Greward = get_reward(row_Hindex, column_Hindex, row_Gindex, column_Gindex)
            old_q_Hvalue = q_Hvalues[old_Hrow_index, old_Hcolumn_index, old_Grow_index, old_Gcolumn_index, Haction_index]
            Htemporal_difference = Hreward + (discount_factor * np.max(q_Hvalues[row_Hindex, column_Hindex, row_Gindex, column_Gindex])) - old_q_Hvalue
            old_q_Gvalue = q_Gvalues[old_Hrow_index, old_Hcolumn_index, old_Grow_index, old_Gcolumn_index, Gaction_index]
            Gtemporal_difference = Greward + (discount_factor * np.max(q_Gvalues[row_Hindex, column_Hindex, row_Gindex, column_Gindex])) - old_q_Gvalue
            
            #update the Q-value for the previous state and action pair
            new_q_Hvalue = old_q_Hvalue + (learning_rate * Htemporal_difference)
            q_Hvalues[old_Hrow_index, old_Hcolumn_index, old_Grow_index, old_Gcolumn_index, Haction_index] = new_q_Hvalue
            new_q_Gvalue = old_q_Gvalue + (learning_rate * Gtemporal_difference)
            q_Gvalues[old_Hrow_index, old_Hcolumn_index, old_Grow_index, old_Gcolumn_index, Gaction_index] = new_q_Gvalue
            
    print('Training complete!')
    print('Episodes:', episodes)



#plot the environment and run a simulation

nrows, ncols = environment_rows,environment_columns

image = np.copy(rewards)
image[environment_rows-1, environment_columns//2] = 100

#row_labels = range(nrows)
#col_labels = range(ncols)
plt.matshow(image, cmap=matplotlib.cm.gist_ncar)
plt.xticks(range(0))#, col_labels)
plt.yticks(range(0))#, row_labels)
plt.show()

def simulation():
    start_Hrow_index, start_Hcolumn_index, start_Grow_index, start_Gcolumn_index = get_starting_location()
    path, t = start_hunting(start_Hrow_index, start_Hcolumn_index, start_Grow_index, start_Gcolumn_index,50, 0.95,0.95)
    image_simu = np.copy(image)
    image_simu[start_Hrow_index, start_Hcolumn_index]= 50
    image_simu[start_Grow_index, start_Gcolumn_index] = -50
    plt.matshow(image_simu, cmap=matplotlib.cm.gist_ncar)
    plt.xticks(range(0))#, col_labels)
    plt.yticks(range(0))#, row_labels)
    plt.show()
    time.sleep(1)
    for i in range(len(path)):
        image_simu = np.copy(image)
        pos_Hrow, pos_Hcol, pos_Grow, pos_Gcol = path[i]
        image_simu[pos_Hrow, pos_Hcol] = 50
        image_simu[pos_Grow, pos_Gcol] = -50
        
        #row_labels = range(nrows)
        #col_labels = range(ncols)
        plt.matshow(image_simu, cmap=matplotlib.cm.gist_ncar)
        plt.xticks(range(0))#, col_labels)
        plt.yticks(range(0))#, row_labels)
        plt.show()
        time.sleep(0.3)
 
#Train the Q function 100K iterations and make a video
training(100000)
simulation()    